"""Handle RGB objects for gradients."""


class RGB:
    """RGB Class."""
    def __init__(self, red: int, green: int, blue: int) -> None:
        """Instantiate the class.

        Args:
            red: red component of RGB color, 0-255
            green: green component of RGB color, 0-255
            blue: blue component of RGB color, 0-255
        """
        self._red = self._verify(red)
        self._green = self._verify(green)
        self._blue = self._verify(blue)

    @staticmethod
    def _verify(color: int) -> int:
        """Verify a given color's value is within acceptable range.

        Args:
            color: value of component RGB color, 0-255
        """
        if not isinstance(color, int):
            raise TypeError("color must be of type 'int'")

        if not ((color >= 0) and (color <= 255)):
            raise ValueError("color must be in range of [0, 255]")

        return color

    # getters
    @property
    def red(self) -> int:
        """Get red component value as int."""
        return self._red

    @property
    def green(self) -> int:
        """Get green component value as int."""
        return self._green

    @property
    def blue(self) -> int:
        """Get blue component value as int."""
        return self._blue



"""Gradient."""
import warnings
from typing import Optional, Union

import numpy
import pandas

import gradpyent.library.formats as formats
from gradpyent.library.rgb import RGB


class Gradient:
    """Gradient class."""

    def __init__(self, gradient_start: Union[RGB, list, tuple, str] = RGB(0, 0, 0),
                 gradient_end: Union[RGB, list, tuple, str] = RGB(255, 255, 255),
                 opacity: Optional[float] = 1.0):
        """Instantiate class.

        Args:
            gradient_start: Gradient 'start' color
            gradient_end: Gradient 'end' color
            opacity: Opacity (for use with KML format)
        """
        self.gradient_start = gradient_start
        self.gradient_end = gradient_end
        self.opacity = opacity

        warnings.formatwarning = self._format_warning

    @staticmethod
    def _format_warning(message: str, category, filename: str, lineno: int, line='') -> str:
        """Create nice looking warnings.

        Args:
            message: The warning message
            category: Category of warning
            filename: File warning is from
            lineno: Line number generating warning
            line: Line
        """
        return f'{filename}:{lineno}\n{category.__name__}: {message}: {line if line is not None else ""}\n'

    # 'public' methods
    def get_gradient_series(self, series: Union[numpy.array, pandas.Series, list], fmt: Optional[str] = None,
                            default: Optional[str] = 'black', force_rescale: Optional[bool] = False) -> list:
        """Create the gradient series.

        Args:
            series: Input series to create color gradient from
            fmt: Desired output format
            default: Default color for nulls
            force_rescale: Force rescale (even if not needed)
        """
        default_rgb = formats.get_verified_color(default)
        lst_colors = []

        # in case numpy.array or list was passed, turn it into a series (also works on an existing series)
        series = pandas.Series(series)

        if series.dtype == 'O':
            warnings.warn("series dtype is object and will be factorized", UserWarning)
            series = pandas.Series(pandas.factorize(series)[0])

        if (series.max() > 1) | (series.min() < 0 | force_rescale):
            warnings.warn("series was out of bounds and will be scaled", UserWarning)
            series = (series - series.min()) / (series.max() - series.min())

        for item in series.values:
            # hacky workaround for numpy float values of 1
            item = min(item, 1)

            # set the gradient color to the list
            if numpy.isnan(item):
                lst_colors.append(formats.format_color(default_rgb, fmt=fmt, opacity=self.opacity))
            else:
                lst_colors.append(self._get_color_from_gradient(item, fmt=fmt))

        return lst_colors

    @property
    def gradient_start(self):
        """Return the gradient starting value as an RGB object."""
        return self._gradient_start

    @gradient_start.setter
    def gradient_start(self, value: Union[RGB, list, tuple, str]):
        """Set the gradient start color.

        Args:
            value: Color
        """
        self._gradient_start = formats.get_verified_color(value)

    @property
    def gradient_end(self):
        """Return the gradient ending value as an RGB object."""
        return self._gradient_end

    @gradient_end.setter
    def gradient_end(self, value: Union[RGB, list, tuple, str]):
        """Set the gradient end color.

        Args:
            value: Color
        """
        self._gradient_end = formats.get_verified_color(value)

    @property
    def opacity(self):
        """Return the opacity value as a float."""
        return self._opacity

    @opacity.setter
    def opacity(self, opacity: float):
        """Set the alhpa (for KML).

        Args:
            opacity: Opacity 0-1
        """
        if (opacity < 0) | (opacity > 1):
            raise ValueError('Opacity value must be in range 0-1 (decimal)')

        # set the property
        self._opacity = opacity
        # hex can be calculated thusly: "hex": format(int(self.o * 255), '02x')

    def set_gradient_bounds(self, gradient_start: Union[RGB, list, tuple, str],
                            gradient_end: Union[RGB, list, tuple, str]):
        """Set gradient start and end colors.

        Args:
            gradient_start: Start color
            gradient_end: End color
        """
        self.gradient_start = gradient_start
        self.gradient_end = gradient_end

    def _get_color_from_gradient(self, percent_gradient: float, fmt: Optional[str] = None) -> Union[str, tuple]:
        """Get gradient color for the given value.

        Args:
            percent_gradient: Percent of gradient range
            fmt: Color output format
        """

        def calc_percent_color(start: int, end: int, percent: float):
            return int(start - ((start - end) * percent))

        if not (percent_gradient >= 0) and (percent_gradient <= 1):
            raise ValueError('Gradient percentage must be 0-1')

        rgb = RGB(
            calc_percent_color(self._gradient_start.red, self._gradient_end.red, percent_gradient),
            calc_percent_color(self._gradient_start.green, self._gradient_end.green, percent_gradient),
            calc_percent_color(self._gradient_start.blue, self._gradient_end.blue, percent_gradient)
        )

        return formats.format_color(rgb, fmt, opacity=self.opacity)



"""Output formatting of colors."""
import re
from collections.abc import Sequence
from functools import singledispatch
from typing import Optional, Tuple, Union

from gradpyent.library.colors import known_colors
from gradpyent.library.rgb import RGB


@singledispatch
def get_verified_color(arg):
    """Dispatch calls to convert input color to RGB.

    Args:
        arg: Color
    """
    raise ValueError(f'{arg} is not a known color value')


@get_verified_color.register
def _(arg: RGB):
    return arg


@get_verified_color.register
def _(arg: Sequence):
    # attempt to convert the given value to RGB format
    return RGB(*arg)


@get_verified_color.register
def _(arg: str):
    if arg in known_colors:
        # if value is in colors, it is a known color, and we can use it as-is
        verified_color = known_colors[arg]
    elif arg[0] == "#":
        # if value starts with '#' it's likely either html or kml fmt
        if len(arg) == 7:
            verified_color = _get_rgb_from_html(arg)
        elif len(arg) == 9:
            verified_color = _get_rgb_from_kml(arg)
        else:
            raise ValueError(f'Unknown string "{arg}" does not match known color, html, or kml type')
    else:
        raise ValueError("Color", arg, " does not exist")

    return verified_color


def _verify_two_char_hex(code: str) -> bool:
    """Verify if a string is a two character hex code.

    Args:
        code: hex string

    Returns:
        True if matched, else False
    """
    return False if (re.search(r'[\dA-Fa-f]{2}', code) is None) else True


def _get_rgb_from_html(code: str) -> RGB:
    """Return RGB object from HTML color code.

    Args:
        code: HTML color code

    Notes:
        HTML color codes are rrggbb
    """
    code = code.lower().replace('#', '')

    red = code[0:2]
    green = code[2:4]
    blue = code[4:6]

    for hex_value in [red, green, blue]:
        if not _verify_two_char_hex(hex_value):
            raise ValueError(f"Expected two-character hex value but got {hex_value}")

    return RGB(
        int(red, 16),    # hex to dec for red
        int(green, 16),  # hex to dec for green
        int(blue, 16)    # hex to dec for blue
    )


def _get_rgb_from_kml(code: str) -> RGB:
    """Return RGB object from KML color code.

    Args:
        code: KML color code

    Notes:
        The first two characters of KML are the alpha (opacity)
        int(code[0:2], 16) / 255
        https://developers.google.com/kml/documentation/kmlreference#color
        The order of expression is aabbggrr
    """
    code = code.lower().replace('#', '')

    # alpha = code[0:2]
    blue = code[2:4]
    green = code[4:6]
    red = code[6:8]

    for hex_value in [red, green, blue]:
        if not _verify_two_char_hex(hex_value):
            raise ValueError(f"Expected two-character hex value but got {hex_value}")

    return RGB(
        int(red, 16),  # hex to dec for red
        int(green, 16),  # hex to dec for green
        int(blue, 16)  # hex to dec for blue
    )


def _get_kml_from_rgb(rgb: RGB, opacity: Optional[float] = 1.0) -> str:
    # return color, formatted for KML with transparency #TTBBGGRR
    """Convert RGB to KML format.

    Args:
        rgb: Color
        opacity: Optionally include opacity, 0-1
    """
    return f'{format(int(opacity * 255), "02x")}{format(int(rgb.blue), "02x")}' \
           f'{format(int(rgb.green), "02x")}{format(int(rgb.red), "02x")}'


def _get_html_from_rgb(rgb: RGB) -> str:
    # return color, formatted for HTML #RRGGBB
    """Convert RGB to HTML format.

    Args:
        rgb: Color
    """
    return f'#{format(int(rgb.red), "02x")}{format(int(rgb.green), "02x")}{format(int(rgb.blue), "02x")}'


def format_color(
        rgb: RGB, fmt: Optional[str] = None, opacity: Optional[float] = None) -> Union[str, Tuple[int, int, int]]:
    """Format output to desired style.

    Args:
        rgb: The RGB object to convert to a different format
        fmt: Desired format
        opacity: If fmt is 'kml' an optional opacity 0-1 can be passed
    """
    if fmt == 'kml':
        formatted_color = _get_kml_from_rgb(rgb, opacity)
    elif fmt == 'html':
        formatted_color = _get_html_from_rgb(rgb)
    elif fmt == 'rgb':
        formatted_color = (rgb.red, rgb.green, rgb.blue)
    else:
        formatted_color = (rgb.red, rgb.green, rgb.blue)

    return formatted_color
